from importlib.util import find_spec
from math import ceil
from multiprocessing import get_context
from os import PathLike
from typing import Literal
from warnings import warn

import numpy as np
import torch
from matplotlib import pyplot as plt
from torch import nn

from mipcandy.common import Normalize, ColorizeLabel
from mipcandy.data.geometric import ensure_num_dimensions


def auto_convert(image: torch.Tensor) -> torch.Tensor:
    return (image * 255 if 0 <= image.min() < image.max() <= 1 else Normalize(domain=(0, 255))(image)).int()


def visualize2d(image: torch.Tensor, *, title: str | None = None, cmap: str = "gray",
                blocking: bool = False, screenshot_as: str | PathLike[str] | None = None) -> None:
    image = image.detach().cpu()
    if image.ndim < 2:
        raise ValueError(f"`image` must have at least 2 dimensions, got {image.shape}")
    if image.ndim > 3:
        image = ensure_num_dimensions(image, 3)
    if image.ndim == 3:
        if image.shape[0] == 1:
            image = image.squeeze(0)
        else:
            image = image.permute(1, 2, 0)
    image = auto_convert(image)
    plt.imshow(image.numpy(), cmap, vmin=0, vmax=255)
    plt.title(title)
    plt.axis("off")
    if screenshot_as:
        plt.savefig(screenshot_as)
        if blocking:
            plt.close()
            return
    plt.show(block=blocking)


def _visualize3d_with_pyvista(image: np.ndarray, title: str | None, cmap: str,
                              screenshot_as: str | PathLike[str] | None) -> None:
    from pyvista import Plotter
    p = Plotter(title=title)
    p.add_volume(image, cmap=cmap)
    if screenshot_as:
        p.off_screen = True
        p.screenshot(screenshot_as)
    else:
        p.show()


def visualize3d(image: torch.Tensor, *, title: str | None = None, cmap: str = "gray", max_volume: int = 1e6,
                backend: Literal["auto", "matplotlib", "pyvista"] = "auto", blocking: bool = False,
                screenshot_as: str | PathLike[str] | None = None) -> None:
    image = image.detach().float().cpu()
    if image.ndim < 3:
        raise ValueError(f"`image` must have at least 3 dimensions, got {image.shape}")
    if image.ndim > 3:
        image = ensure_num_dimensions(image, 3)
    d, h, w = image.shape
    total = d * h * w
    ratio = int(ceil((total / max_volume) ** (1 / 3))) if total > max_volume else 1
    if ratio > 1:
        image = ensure_num_dimensions(nn.functional.avg_pool3d(ensure_num_dimensions(image, 5), kernel_size=ratio,
                                                               stride=ratio, ceil_mode=True), 3)
    image /= image.max()
    image = image.numpy()
    if backend == "auto":
        backend = "pyvista" if find_spec("pyvista") else "matplotlib"
    match backend:
        case "matplotlib":
            warn("Using Matplotlib for 3D visualization is inefficient and inaccurate, consider using PyVista")
            face_colors = getattr(plt.cm, cmap)(image)
            face_colors[..., 3] = image * (image > 0)
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.voxels(image, facecolors=face_colors)
            ax.set_title(title)
            if screenshot_as:
                fig.savefig(screenshot_as)
                if blocking:
                    plt.close()
                    return
            plt.show(block=blocking)
        case "pyvista":
            image = image.transpose(1, 2, 0)
            if blocking:
                return _visualize3d_with_pyvista(image, title, cmap, screenshot_as)
            ctx = get_context("spawn")
            return ctx.Process(target=_visualize3d_with_pyvista, args=(image, title, cmap, screenshot_as),
                               daemon=False).start()


def overlay(image: torch.Tensor, label: torch.Tensor, *, max_label_opacity: float = .5,
            label_colorizer: ColorizeLabel | None = ColorizeLabel()) -> torch.Tensor:
    if image.ndim < 2 or label.ndim < 2:
        raise ValueError("Only 2D images can be overlaid")
    image = ensure_num_dimensions(image, 3)
    label = ensure_num_dimensions(label, 2)
    image = auto_convert(image)
    if image.shape[0] == 1:
        image = image.repeat(3, 1, 1)
    image_c, image_shape = image.shape[0], image.shape[1:]
    label_shape = label.shape
    if image_shape != label_shape:
        raise ValueError(f"Unmatched shapes {image_shape} and {label_shape}")
    alpha = (label > 0).int()
    if label_colorizer:
        label = label_colorizer(label)
        if label.shape[0] == 4:
            alpha = label[-1]
            label = label[:-1]
    elif label.shape[0] == 1:
        label = label.repeat(3, 1, 1)
    if not (image_c == label.shape[0] == 3):
        raise ValueError("Unsupported number of channels")
    if alpha.max() > 0:
        alpha = alpha * max_label_opacity / alpha.max()
    return image * (1 - alpha) + label * alpha



from typing import Literal

import torch
from torch import nn


def convert_ids_to_logits(ids: torch.Tensor, d: Literal[1, 2, 3], num_classes: int) -> torch.Tensor:
    if ids.dtype != torch.int or ids.min() < 0:
        raise TypeError("`ids` should be positive integers")
    d += 1
    if ids.ndim != d:
        if ids.ndim == d + 1 and ids.shape[1] == 1:
            ids = ids.squeeze(1)
        else:
            raise ValueError(f"`ids` should be {d} dimensional or {d + 1} dimensional with single channel")
    return nn.functional.one_hot(ids.long(), num_classes).movedim(-1, 1).float()


def convert_logits_to_ids(logits: torch.Tensor, *, channel_dim: int = 1) -> torch.Tensor:
    return logits.max(channel_dim).indices.int()



from typing import Any, Generator, Self

from torch import nn

from mipcandy.types import Device


def batch_int_multiply(f: float, *n: int) -> Generator[int, None, None]:
    for i in n:
        r = i * f
        if not r.is_integer():
            raise ValueError(f"Inequivalent conversion")
        yield int(r)


def batch_int_divide(f: float, *n: int) -> Generator[int, None, None]:
    return batch_int_multiply(1 / f, *n)


class LayerT(object):
    def __init__(self, m: type[nn.Module], **kwargs) -> None:
        self.m: type[nn.Module] = m
        self.kwargs: dict[str, Any] = kwargs

    def update(self, *, must_exist: bool = True, **kwargs) -> Self:
        for k, v in kwargs.items():
            if not must_exist or k in self.kwargs:
                self.kwargs[k] = v
        return self

    def assemble(self, *args, **kwargs) -> nn.Module:
        self_kwargs = self.kwargs.copy()
        for k, v in self_kwargs.items():
            if isinstance(v, str) and v in kwargs:
                self_kwargs[k] = kwargs.pop(v)
        return self.m(*args, **self_kwargs, **kwargs)


class HasDevice(object):
    def __init__(self, device: Device) -> None:
        self._device: Device = device


class WithPaddingModule(HasDevice):
    def __init__(self, device: Device) -> None:
        super().__init__(device)
        self._padding_module: nn.Module | None = None
        self._restoring_module: nn.Module | None = None
        self._padding_module_built: bool = False

    def build_padding_module(self) -> nn.Module | None:
        return None

    def build_restoring_module(self, padding_module: nn.Module | None) -> nn.Module | None:
        return None

    def _lazy_load_padding_module(self) -> None:
        if self._padding_module_built:
            return
        self._padding_module = self.build_padding_module()
        if self._padding_module:
            self._padding_module = self._padding_module.to(self._device)
        self._restoring_module = self.build_restoring_module(self._padding_module)
        if self._restoring_module:
            self._restoring_module = self._restoring_module.to(self._device)
        self._padding_module_built = True

    def get_padding_module(self) -> nn.Module | None:
        self._lazy_load_padding_module()
        return self._padding_module

    def get_restoring_module(self) -> nn.Module | None:
        self._lazy_load_padding_module()
        return self._restoring_module



from typing import override

import torch
from torch import nn

from mipcandy.layer import LayerT


class AbstractConvBlock(nn.Module):
    def __init__(self, in_ch: int, out_ch: int, kernel_size: int, *, stride: int = 1, padding: int = 0,
                 dilation: int = 1, groups: int = 1, bias: bool = True, padding_mode: str = "zeros",
                 conv: LayerT = ..., norm: LayerT = ..., act: LayerT = ...) -> None:
        super().__init__()
        self.conv: nn.Module = conv.assemble(in_ch, out_ch, kernel_size, stride, padding, dilation, groups, bias,
                                             padding_mode)
        self.norm: nn.Module = norm.assemble(in_ch=out_ch)
        self.act: nn.Module = act.assemble()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.act(self.norm(self.conv(x)))


def _conv_block(default_conv: LayerT, default_norm: LayerT, default_act: LayerT) -> type[AbstractConvBlock]:
    class ConvBlock(AbstractConvBlock):
        def __init__(self, *args, **kwargs) -> None:
            if "conv" not in kwargs:
                kwargs["conv"] = default_conv
            if "norm" not in kwargs:
                kwargs["norm"] = default_norm
            if "act" not in kwargs:
                kwargs["act"] = default_act
            super().__init__(*args, **kwargs)

    return ConvBlock


ConvBlock2d: type[AbstractConvBlock] = _conv_block(
    LayerT(nn.Conv2d), LayerT(nn.BatchNorm2d, num_features="in_ch"), LayerT(nn.ReLU, inplace=True)
)

ConvBlock3d: type[AbstractConvBlock] = _conv_block(
    LayerT(nn.Conv3d), LayerT(nn.BatchNorm3d, num_features="in_ch"), LayerT(nn.ReLU, inplace=True)
)


class WSConv2d(nn.Conv2d):
    @override
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        w = self.weight
        w = (w - w.mean(dim=(1, 2, 3), keepdim=True)) / (w.std(dim=(1, 2, 3), keepdim=True) + 1e-5)
        return nn.functional.conv2d(x, w, self.bias, self.stride, self.padding, self.dilation, self.groups)


class WSConv3d(nn.Conv3d):
    @override
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        w = self.weight
        w = (w - w.mean(dim=(1, 2, 3, 4), keepdim=True)) / (w.std(dim=(1, 2, 3, 4), keepdim=True) + 1e-5)
        return nn.functional.conv3d(x, w, self.bias, self.stride, self.padding, self.dilation, self.groups)



from random import seed as random_seed, randint
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from hashlib import md5
from os import PathLike, urandom, makedirs, environ
from os.path import exists
from shutil import copy
from threading import Lock
from time import time
from typing import Sequence, override

import numpy as np
import torch
from matplotlib import pyplot as plt
from pandas import DataFrame
from rich.progress import Progress, SpinnerColumn
from torch import nn, optim
from torch.utils.data import DataLoader

from mipcandy.common import Pad2d, Pad3d, quotient_regression, quotient_derivative, quotient_bounds
from mipcandy.config import load_settings, load_secrets
from mipcandy.frontend import Frontend
from mipcandy.layer import WithPaddingModule
from mipcandy.sanity_check import sanity_check
from mipcandy.sliding_window import SWMetadata, SlidingWindow
from mipcandy.types import Params, Setting, Device


def try_append(new: float, to: dict[str, list[float]], key: str) -> None:
    if key in to:
        to[key].append(new)
    else:
        to[key] = [new]


def try_append_all(new: dict[str, float], to: dict[str, list[float]]) -> None:
    for key, value in new.items():
        try_append(value, to, key)


def _is_reserved(metric: str) -> bool:
    return metric == "learning rate" or metric == "epoch duration" or metric.startswith("val")


@dataclass
class TrainerToolbox(object):
    model: nn.Module
    optimizer: optim.Optimizer
    scheduler: optim.lr_scheduler.LRScheduler
    criterion: nn.Module
    ema: optim.swa_utils.AveragedModel | None = None


class Trainer(WithPaddingModule, metaclass=ABCMeta):
    def __init__(self, trainer_folder: str | PathLike[str], dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]],
                 validation_dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]], *,
                 device: Device = "cpu") -> None:
        super().__init__(device)
        self._trainer_folder: str = trainer_folder
        self._trainer_variant: str = self.__class__.__name__
        self._experiment_id: str = "tbd"
        self._dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]] = dataloader
        self._validation_dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]] = validation_dataloader
        self._metrics: dict[str, list[float]] = {}
        self._epoch_metrics: dict[str, list[float]] = {}
        self._best_score: float = float("-inf")
        self._worst_case: tuple[torch.Tensor, torch.Tensor, torch.Tensor] | None = None
        self._frontend: Frontend = Frontend({})
        self._lock: Lock = Lock()

    def set_frontend(self, frontend: type[Frontend], *, path_to_secrets: str | PathLike[str] | None = None) -> None:
        self._frontend = frontend(load_secrets(path=path_to_secrets) if path_to_secrets else load_secrets())

    def initialized(self) -> bool:
        return self._experiment_id != "tbd"

    def experiment_folder(self) -> str:
        return f"{self._trainer_folder}/{self._trainer_variant}/{self._experiment_id}"

    def allocate_experiment_folder(self) -> str:
        self._experiment_id = datetime.now().strftime("%Y%m%d-%H-") + md5(urandom(8)).hexdigest()[:4]
        experiment_folder = self.experiment_folder()
        return self.allocate_experiment_folder() if exists(experiment_folder) else experiment_folder

    def init_experiment(self) -> None:
        if self.initialized():
            raise RuntimeError("Experiment already initialized")
        makedirs(self._trainer_folder, exist_ok=True)
        experiment_folder = self.allocate_experiment_folder()
        makedirs(experiment_folder)
        t = datetime.now()
        with open(f"{experiment_folder}/logs.txt", "w") as f:
            f.write(f"File created by FightTumor, copyright (C) {t.year} Project Neura. All rights reserved\n")
        self.log(f"Experiment (ID {self._experiment_id}) created at {t}")
        self.log(f"Trainer: {self.__class__.__name__}")

    def log(self, msg: str, *, console: bool = True) -> None:
        msg = f"[{datetime.now()}] {msg}"
        if self.initialized():
            with open(f"{self.experiment_folder()}/logs.txt", "a") as f:
                f.write(f"{msg}\n")
        if console:
            with self._lock:
                print(msg)

    def record(self, metric: str, value: float) -> None:
        try_append(value, self._epoch_metrics, metric)

    def _record(self, metric: str, value: float) -> None:
        try_append(value, self._metrics, metric)

    def record_all(self, metrics: dict[str, float]) -> None:
        try_append_all(metrics, self._epoch_metrics)

    def _bump_metrics(self) -> None:
        for metric, values in self._epoch_metrics.items():
            epoch_overall = sum(values) / len(values)
            try_append(epoch_overall, self._metrics, metric)
        self._epoch_metrics.clear()

    def save_metrics(self) -> None:
        df = DataFrame(self._metrics)
        df.index = range(1, len(df) + 1)
        df.index.name = "epoch"
        df.to_csv(f"{self.experiment_folder()}/metrics.csv")

    def save_metric_curve(self, name: str, values: Sequence[float]) -> None:
        name = name.capitalize()
        plt.plot(values)
        plt.title(f"{name} over Epoch")
        plt.xlabel("Epoch")
        plt.ylabel(name)
        plt.grid()
        plt.savefig(f"{self.experiment_folder()}/{name.lower()}.png")
        plt.close()

    def save_metric_curve_combo(self, metrics: dict[str, Sequence[float]], *, title: str = "All Metrics") -> None:
        for name, values in metrics.items():
            plt.plot(values, label=name.capitalize())
        plt.title(title)
        plt.xlabel("Epoch")
        plt.legend()
        plt.grid()
        plt.savefig(f"{self.experiment_folder()}/{title.lower()}.png")
        plt.close()

    def save_metric_curves(self, *, names: Sequence[str] | None = None) -> None:
        if names is None:
            for name, values in self._metrics.items():
                self.save_metric_curve(name, values)
        else:
            for name in names:
                self.save_metric_curve(name, self._metrics[name])

    def save_progress(self, *, names: Sequence[str] = ("combined loss", "val score")) -> None:
        self.save_metric_curve_combo({name: self._metrics[name] for name in names}, title="Progress")

    def save_preview(self, image: torch.Tensor, label: torch.Tensor, mask: torch.Tensor, *,
                     quality: float = .75) -> None:
        ...

    @abstractmethod
    def build_network(self, example_shape: tuple[int, ...]) -> nn.Module:
        raise NotImplementedError

    @abstractmethod
    def build_optimizer(self, params: Params) -> optim.Optimizer:
        raise NotImplementedError

    @abstractmethod
    def build_scheduler(self, optimizer: optim.Optimizer, num_epochs: int) -> optim.lr_scheduler.LRScheduler:
        raise NotImplementedError

    @abstractmethod
    def build_criterion(self) -> nn.Module:
        raise NotImplementedError

    @abstractmethod
    def backward(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        raise NotImplementedError

    def train_batch(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        toolbox.optimizer.zero_grad()
        loss, metrics = self.backward(images, labels, toolbox)
        toolbox.optimizer.step()
        toolbox.scheduler.step()
        if toolbox.ema:
            toolbox.ema.update_parameters(toolbox.model)
        return loss, metrics

    def train_epoch(self, epoch: int, toolbox: TrainerToolbox) -> None:
        toolbox.model.train()
        if toolbox.ema:
            toolbox.ema.train()
        with Progress(*Progress.get_default_columns(), SpinnerColumn()) as progress:
            epoch_prog = progress.add_task(f"Epoch {epoch}", total=len(self._dataloader))
            for images, labels in self._dataloader:
                images, labels = images.to(self._device), labels.to(self._device)
                padding_module = self.get_padding_module()
                if padding_module:
                    images, labels = padding_module(images), padding_module(labels)
                loss, metrics = self.train_batch(images, labels, toolbox)
                self.record("combined loss", loss)
                self.record_all(metrics)
                progress.update(epoch_prog, advance=1, description=f"Training epoch {epoch} ({loss:.4f})")
        self._bump_metrics()

    @abstractmethod
    def validate_case(self, image: torch.Tensor, label: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float], torch.Tensor]:
        raise NotImplementedError

    def validate(self, toolbox: TrainerToolbox) -> tuple[float, dict[str, list[float]]]:
        if self._validation_dataloader.batch_size != 1:
            raise RuntimeError("Validation dataloader should have batch size 1")
        toolbox.model.eval()
        if toolbox.ema:
            toolbox.ema.eval()
        score = 0
        worst_score = float("+inf")
        metrics = {}
        num_cases = len(self._validation_dataloader)
        with Progress(*Progress.get_default_columns(), SpinnerColumn()) as progress:
            val_prog = progress.add_task(f"Validating", total=num_cases)
            for image, label in self._validation_dataloader:
                image, label = image.to(self._device), label.to(self._device)
                padding_module = self.get_padding_module()
                if padding_module:
                    image, label = padding_module(image), padding_module(label)
                image, label = image.squeeze(0), label.squeeze(0)
                case_score, case_metrics, mask = self.validate_case(image, label, toolbox)
                score += case_score
                if case_score < worst_score:
                    self._worst_case = (image, label, mask)
                    worst_score = case_score
                try_append_all(case_metrics, metrics)
                progress.update(val_prog, advance=1, description=f"Validating ({case_score:.4f})")
        return score / num_cases, metrics

    def predict_maximum_validation_score(self, num_epochs: int, *, degree: int = 5) -> tuple[int, float]:
        val_scores = np.array(self._metrics["val score"])
        a, b = quotient_regression(np.arange(len(val_scores)), val_scores, degree, degree)
        da, db = quotient_derivative(a, b)
        max_roc = float(da[0] / db[0])
        max_val_score = float(a[0] / b[0])
        bounds = quotient_bounds(a, b, None, max_val_score * (1 - max_roc), x_start=0, x_stop=num_epochs, x_step=1)
        return round(bounds[1]) + 1, max_val_score if bounds else (0, 0)

    def set_seed(self, seed: int) -> None:
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        random_seed(seed)
        np.random.seed(seed)
        environ['PYTHONHASHSEED'] = str(seed)
        if self.initialized():
            self.log(f"Set to manual seed {seed}")

    @staticmethod
    def filter_train_params(**kwargs) -> dict[str, Setting]:
        return {k: v for k, v in kwargs.items() if k in (
            "note", "num_checkpoints", "ema", "seed", "early_stop_tolerance", "val_score_prediction",
            "val_score_prediction_degree", "save_preview", "preview_quality"
        )}

    def train_with_settings(self, num_epochs: int, **kwargs) -> None:
        settings = self.filter_train_params(**load_settings())
        settings.update(kwargs)
        self.train(num_epochs, **settings)

    def train(self, num_epochs: int, *, note: str = "", num_checkpoints: int = 5, ema: bool = True,
              seed: int | None = None, early_stop_tolerance: int = 5, val_score_prediction: bool = True,
              val_score_prediction_degree: int = 5, save_preview: bool = True, preview_quality: float = .75) -> None:
        self.init_experiment()
        if note:
            self.log(f"Note: {note}")
        if seed is None:
            seed = randint(0, 100)
        self.set_seed(seed)
        example_input = self._dataloader.dataset[0][0].to(self._device).unsqueeze(0)
        padding_module = self.get_padding_module()
        if padding_module:
            example_input = padding_module(example_input)
        example_shape = tuple(example_input.shape[1:])
        self.log(f"Example input shape: {example_shape}")
        model = self.build_network(example_shape).to(self._device)
        model_name = model.__class__.__name__
        sanity_check_result = sanity_check(model, example_shape, device=self._device)
        self.log(f"Model: {model_name}")
        self.log(str(sanity_check_result))
        self.log(f"Example output shape: {tuple(sanity_check_result.output.shape)}")
        optimizer = self.build_optimizer(model.parameters())
        scheduler = self.build_scheduler(optimizer, num_epochs)
        criterion = self.build_criterion().to(self._device)
        toolbox = TrainerToolbox(model, optimizer, scheduler, criterion)
        if ema:
            toolbox.ema = optim.swa_utils.AveragedModel(model)
        checkpoint_path = lambda v: f"{self.experiment_folder()}/checkpoint_{v}.pth"
        es_tolerance = early_stop_tolerance
        self._frontend.on_experiment_created(self._experiment_id, self._trainer_variant, model_name, note,
                                             sanity_check_result.num_macs, sanity_check_result.num_params, num_epochs,
                                             early_stop_tolerance)
        try:
            for epoch in range(1, num_epochs + 1):
                if early_stop_tolerance == -1:
                    epoch -= 1
                    self.log(f"Early stopping triggered because the validation score has not improved for {
                    es_tolerance} epochs")
                    break
                # Training
                t0 = time()
                self.train_epoch(epoch, toolbox)
                lr = scheduler.get_last_lr()[0]
                self._record("learning rate", lr)
                self.log(f"Learning rate: {lr}")
                for metric, values in self._metrics.items():
                    if _is_reserved(metric):
                        continue
                    msg = f"Training {metric}: {values[-1]:.4f}"
                    if epoch > 1:
                        msg += f" ({values[-1] - values[-2]:+.4f})"
                    self.log(msg)
                torch.save(model.state_dict(), checkpoint_path("latest"))
                if epoch % (num_epochs / num_checkpoints) == 0:
                    copy(checkpoint_path("latest"), checkpoint_path(epoch))
                    self.log(f"Epoch {epoch} checkpoint saved")
                self.log(f"Epoch {epoch} training completed in {time() - t0:.1f} seconds")
                # Validation
                score, metrics = self.validate(toolbox)
                self._record("val score", score)
                msg = f"Validation score: {score:.4f}"
                if epoch > 1:
                    msg += f" ({score - self._metrics["val score"][-2]:+.4f})"
                self.log(msg)
                if val_score_prediction and epoch > val_score_prediction_degree:
                    target_epoch, max_score = self.predict_maximum_validation_score(
                        num_epochs, degree=val_score_prediction_degree
                    )
                    self.log(f"Maximum validation score {max_score:.4f} predicted at epoch {target_epoch}")
                    epoch_durations = self._metrics["epoch duration"]
                    etc = sum(epoch_durations) * (target_epoch - epoch) / len(epoch_durations)
                    self.log(f"Estimated time of completion in {etc:.1f} seconds: {datetime.fromtimestamp(
                        time() + etc):%H:%M:%S}")
                for metric, values in metrics.items():
                    a, b, c = min(values), sum(values) / len(values), max(values)
                    self.log(f"Validation {metric}: {b:.4f} @[{a:.4f}, {c:.4f}]")
                if score > self._best_score:
                    copy(checkpoint_path("latest"), checkpoint_path("best"))
                    self.log(f"======== Best checkpoint updated ({self._best_score:.4f} -> {score:.4f}) ========")
                    self._best_score = score
                    early_stop_tolerance = es_tolerance
                    if save_preview:
                        self.save_preview(*self._worst_case, quality=preview_quality)
                else:
                    early_stop_tolerance -= 1
                epoch_duration = time() - t0
                self._record("epoch duration", epoch_duration)
                self.log(f"Epoch {epoch} completed in {epoch_duration:.1f} seconds")
                self.log(f"=============== Best Validation Score {self._best_score:.4f} ===============")
                self.save_metrics()
                self.save_progress()
                self.save_metric_curves()
                self._frontend.on_experiment_updated(self._experiment_id, epoch, self._metrics, early_stop_tolerance)
        except Exception as e:
            self.log("Training interrupted")
            self.log(repr(e))
            self._frontend.on_experiment_interrupted(self._experiment_id, e)
            raise e
        else:
            self.log("Training completed")
            self._frontend.on_experiment_completed(self._experiment_id)

    def __call__(self, *args, **kwargs) -> None:
        self.train(*args, **kwargs)


class SlidingTrainer(Trainer, SlidingWindow, metaclass=ABCMeta):
    @override
    def build_padding_module(self) -> nn.Module | None:
        window_shape = self.get_window_shape()
        return (Pad2d if len(window_shape) == 2 else Pad3d)(window_shape)

    @abstractmethod
    def validate_case_windowed(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox,
                               metadata: SWMetadata) -> tuple[float, dict[str, float], torch.Tensor]:
        raise NotImplementedError

    @override
    def validate_case(self, image: torch.Tensor, label: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float], torch.Tensor]:
        image, label = image.unsqueeze(0), label.unsqueeze(0)
        images, metadata = self.do_sliding_window(image)
        labels, _ = self.do_sliding_window(label)
        loss, metrics, masks = self.validate_case_windowed(images, labels, toolbox, metadata)
        mask = self.revert_sliding_window(masks, metadata)
        return loss, metrics, mask.squeeze(0)

    @abstractmethod
    def backward_windowed(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox,
                          metadata: SWMetadata) -> tuple[float, dict[str, float]]:
        raise NotImplementedError

    @override
    def backward(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        images, metadata = self.do_sliding_window(images)
        labels, _ = self.do_sliding_window(labels)
        return self.backward_windowed(images, labels, toolbox, metadata)



from typing import override

from torch import optim


class AbsoluteLinearLR(optim.lr_scheduler.LRScheduler):
    """
    lr = kx + b
    """
    def __init__(self, optimizer: optim.Optimizer, k: float, b: float, *, min_lr: float = 1e-6,
                 restart: bool = False, last_epoch: int = -1) -> None:
        self._k: float = k
        self._b: float = b
        if min_lr < 0:
            raise ValueError(f"`min_lr` must be positive, but got {min_lr}")
        self._min_lr: float = min_lr
        self._restart: bool = restart
        self._restart_step: int = 0
        super().__init__(optimizer, last_epoch)

    def _interp(self, step: int) -> float:
        step -= self._restart_step
        r = self._k * step + self._b
        if r < self._min_lr:
            if self._restart:
                self._restart_step = step
                return self._interp(step)
            return self._min_lr
        return r

    @override
    def get_lr(self) -> list[float]:
        target = self._interp(self.last_epoch)
        return [target for _ in self.optimizer.param_groups]



from mipcandy.__entry__ import __entry__
from mipcandy.common import *
from mipcandy.config import load_settings, save_settings, load_secrets, save_secrets
from mipcandy.data import *
from mipcandy.evaluation import EvalCase, EvalResult, Evaluator
from mipcandy.frontend import *
from mipcandy.inference import parse_predictant, Predictor
from mipcandy.layer import batch_int_multiply, batch_int_divide, LayerT, HasDevice, WithPaddingModule
from mipcandy.metrics import do_reduction, dice_similarity_coefficient_binary, \
    dice_similarity_coefficient_multiclass, soft_dice_coefficient, accuracy_binary, accuracy_multiclass, \
    precision_binary, precision_multiclass, recall_binary, recall_multiclass, iou_binary, iou_multiclass
from mipcandy.presets import *
from mipcandy.run import config
from mipcandy.sanity_check import num_trainable_params, model_complexity_info, SanityCheckResult, sanity_check
from mipcandy.training import TrainerToolbox, Trainer, SWMetadata, SlidingTrainer
from mipcandy.types import Setting, Settings, Params, Transform, SupportedPredictant, Colormap, Device



from dataclasses import dataclass
from typing import Callable, Sequence, Generator, override

import torch

from mipcandy.data import SupervisedDataset, MergedDataset, Loader, DatasetFromMemory
from mipcandy.inference import parse_predictant, Predictor
from mipcandy.types import SupportedPredictant


@dataclass
class EvalCase(object):
    metrics: dict[str, float]
    mask: torch.Tensor
    label: torch.Tensor
    image: torch.Tensor | None = None
    filename: str | None = None


class EvalResult(Sequence[EvalCase]):
    def __init__(self, metrics: dict[str, list[float]], masks: list[torch.Tensor], labels: list[torch.Tensor], *,
                 images: list[torch.Tensor] | None = None, filenames: list[str] | None = None) -> None:
        if len(masks) != len(labels):
            raise ValueError(f"Unmatched number of masks ({len(masks)}) and labels ({len(labels)})")
        self.metrics: dict[str, list[float]] = metrics
        self.mean_metrics: dict[str, float] = {name: sum(values) / len(values) for name, values in metrics.items()}
        self.images: list[torch.Tensor] | None = images
        self.masks: list[torch.Tensor] = masks
        self.labels: list[torch.Tensor] = labels
        self.filenames: list[str] | None = filenames

    @override
    def __len__(self) -> int:
        return len(self.masks)

    @override
    def __getitem__(self, item: int) -> EvalCase:
        return EvalCase({name: values[item] for name, values in self.metrics.items()}, self.masks[item],
                        self.labels[item], self.images[item] if self.images else None,
                        self.filenames[item] if self.filenames else None)

    def _select(self, metric: str, n: int, descending: bool) -> Generator[EvalCase, None, None]:
        o_values = self.metrics[metric]
        values = o_values.copy()
        values.sort(reverse=descending)
        for value in values[:n]:
            yield self[o_values.index(value)]

    def min(self, metric: str) -> EvalCase:
        return self.min_n(metric, 1)[0]

    def min_n(self, metric: str, n: int) -> tuple[EvalCase, ...]:
        return tuple(self._select(metric, n, False))

    def max(self, metric: str) -> EvalCase:
        return self.max_n(metric, 1)[0]

    def max_n(self, metric: str, n: int) -> tuple[EvalCase, ...]:
        return tuple(self._select(metric, n, True))


class Evaluator(object):
    def __init__(self, *metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]) -> None:
        self._metrics: tuple[Callable[[torch.Tensor, torch.Tensor], torch.Tensor], ...] = metrics

    def _evaluate_dataset(self, x: SupervisedDataset, *, prefilled_masks: list[torch.Tensor] | None = None,
                          prefilled_labels: list[torch.Tensor] | None = None) -> EvalResult:
        metrics = {}
        masks = prefilled_masks if prefilled_masks else []
        labels = prefilled_labels if prefilled_labels else []
        for mask, label in x:
            if not prefilled_masks:
                masks.append(mask)
            if not prefilled_labels:
                labels.append(label)
            for m in self._metrics:
                if m.__name__ not in metrics:
                    metrics[m.__name__] = []
                metrics[m.__name__].append(m(mask, label).item())
        return EvalResult(metrics, masks, labels)

    def evaluate_dataset(self, x: SupervisedDataset) -> EvalResult:
        return self._evaluate_dataset(x)

    def evaluate(self, masks: SupportedPredictant, labels: SupportedPredictant) -> EvalResult:
        masks, filenames = parse_predictant(masks, Loader)
        labels, _ = parse_predictant(labels, Loader, as_label=True)
        r = self._evaluate_dataset(MergedDataset(DatasetFromMemory(masks), DatasetFromMemory(labels)),
                                   prefilled_masks=masks, prefilled_labels=labels)
        r.filenames = filenames
        return r

    def predict_and_evaluate(self, x: SupportedPredictant, labels: SupportedPredictant,
                             predictor: Predictor) -> EvalResult:
        x, filenames = parse_predictant(x, Loader)
        masks = [e.cpu() for e in predictor.predict(x)]
        r = self.evaluate(masks, labels)
        r.images = x
        r.filenames = filenames
        return r



from argparse import ArgumentParser

from mipcandy.run import config


def __entry__() -> None:
    parser = ArgumentParser(prog="MIP Candy CLI", description="MIP Candy Command Line Interface",
                            epilog="GitHub: https://github.com/ProjectNeura/MIPCandy")
    parser.add_argument("-c", "--config", choices=("setting", "secret"), default=None,
                        help="set a configuration such that key=value")
    parser.add_argument("-kv", "--key-value", nargs=2, action="append", default=None, help="define a key-value pair")
    args = parser.parse_args()
    if args.config:
        if not args.key_value:
            raise ValueError("Expected at least one key-value pair")
        for key_value in args.key_value:
            config(args.config, key_value[0], key_value[1])



from datetime import datetime
from typing import override, Literal

from requests import get, post, patch, Response

from mipcandy.frontend.prototype import Frontend
from mipcandy.types import Settings


class NotionFrontend(Frontend):
    def __init__(self, secrets: Settings) -> None:
        super().__init__(secrets)
        self._api_key: str = self.require_nonempty_secret("notion_api_key", require_type=str)
        self._database_id: str = self.require_nonempty_secret("notion_database_id", require_type=str)
        self._headers: dict[str, str] = {
            "Authorization": f"Bearer {self._api_key}",
            "Content-Type": "application/json",
            "Notion-Version": "2022-06-28"
        }
        self._num_epochs: int = 1
        self._early_stop_tolerance: int = -1
        self._start_time: str = ""
        self._page_id: str = ""

    def retrieve_database(self) -> Response:
        return get(f"https://api.notion.com/v1/databases/{self._database_id}", headers=self._headers)

    def query_database(self, *, experiment_id: str | None = None) -> Response:
        json = {"filter": {"property": "Experiment ID", "title": {"equals": experiment_id}}} if experiment_id else None
        return post(f"https://api.notion.com/v1/databases/{self._database_id}/query", json=json, headers=self._headers)

    def select_experiment(self, experiment_id: str) -> str:
        experiments = self.query_database(experiment_id=experiment_id)
        if experiments.status_code != 200:
            raise RuntimeError(f"Failed to query database: {experiments.json()}")
        experiments = experiments.json()["results"]
        if len(experiments) == 1:
            return experiments[0]["id"]
        if len(experiments) > 1:
            raise RuntimeError(f"Found multiple experiments with the same ID {experiment_id}")
        return ""

    def new_experiment(self, experiment_id: str, trainer: str, model: str, note: str, num_macs: float,
                       num_params: float) -> Response:
        self._start_time = datetime.now().astimezone().strftime("%Y-%m-%dT%H:%M:%S.000%z")
        properties = {
            "Experiment ID": {"title": [{"text": {"content": experiment_id}}]},
            "Status": {"status": {"name": "In Progress"}},
            "Progress": {"number": 0},
            "Early Stop": {"number": 1},
            "Trainer": {"select": {"name": trainer}},
            "Model": {"select": {"name": model}},
            "Time": {"date": {"start": self._start_time}},
            "Note": {"rich_text": [{"text": {"content": note}}]},
            "MACs (G)": {"number": round(num_macs, 1)},
            "Params (M)": {"number": round(num_params, 1)},
            "Epoch": {"number": 0},
            "Score": {"number": 0},
        }
        page_id = self.select_experiment(experiment_id)
        if page_id:
            self._page_id = page_id
            return patch(f"https://api.notion.com/v1/pages/{page_id}", json={"properties": properties},
                         headers=self._headers)
        res = post("https://api.notion.com/v1/pages", json={
            "parent": {"database_id": self._database_id},
            "icon": {"external": {"url": "https://www.notion.so/icons/science_gray.svg"}},
            "properties": properties
        }, headers=self._headers)
        self._page_id = res.json()["id"]
        return res

    def update_experiment(self, experiment_id: str, status: Literal["In Progress", "Completed", "Interrupted"],
                          *, epoch: int | None = None, score: float | None = None,
                          early_stop_tolerance: int | None = None, observation: str | None = None) -> Response:
        if not self._page_id:
            raise RuntimeError(f"Experiment {experiment_id} has not been created")
        properties = {"Status": {"status": {"name": status}}}
        if epoch is not None:
            properties["Progress"] = {"number": epoch / self._num_epochs}
            properties["Epoch"] = {"number": epoch}
        if early_stop_tolerance is not None:
            properties["Early Stop"] = {"number": max(early_stop_tolerance, 0) / self._early_stop_tolerance}
        if score is not None:
            properties["Score"] = {"number": round(score, 4)}
        if observation is not None:
            properties["Observation"] = {"rich_text": [{"text": {"content": observation}}]}
        if status == "Completed":
            properties["Progress"] = {"number": 1}
            properties["Time"] = {"date": {"start": self._start_time,
                                           "end": datetime.now().astimezone().strftime("%Y-%m-%dT%H:%M:%S.000%z")}}
        return patch(f"https://api.notion.com/v1/pages/{self._page_id}", json={"properties": properties},
                     headers=self._headers)

    @override
    def on_experiment_created(self, experiment_id: str, trainer: str, model: str, note: str, num_macs: float,
                              num_params: float, num_epochs: int, early_stop_tolerance: int) -> None:
        self._num_epochs = num_epochs
        self._early_stop_tolerance = early_stop_tolerance
        res = self.new_experiment(experiment_id, trainer, model, note, num_macs, num_params)
        if res.status_code != 200:
            raise RuntimeError(f"Failed to create experiment: {res.json()}")

    @override
    def on_experiment_updated(self, experiment_id: str, epoch: int, metrics: dict[str, list[float]],
                              early_stop_tolerance: int) -> None:
        try:
            self.update_experiment(experiment_id, "In Progress", epoch=epoch, score=max(metrics["val score"]),
                                   early_stop_tolerance=early_stop_tolerance)
        except RuntimeError:
            pass

    @override
    def on_experiment_completed(self, experiment_id: str) -> None:
        res = self.update_experiment(experiment_id, "Completed")
        if res.status_code != 200:
            raise RuntimeError(f"Failed to update experiment: {res.json()}")

    @override
    def on_experiment_interrupted(self, experiment_id: str, error: Exception) -> None:
        res = self.update_experiment(experiment_id, "Interrupted", observation=repr(error))
        if res.status_code != 200:
            raise RuntimeError(f"Failed to update experiment: {res.json()}")



from mipcandy.data.convertion import convert_ids_to_logits, convert_logits_to_ids
from mipcandy.data.dataset import Loader, UnsupervisedDataset, SupervisedDataset, DatasetFromMemory, MergedDataset, \
    NNUNetDataset, BinarizedDataset
from mipcandy.data.geometric import ensure_num_dimensions, orthographic_views, aggregate_orthographic_views
from mipcandy.data.io import resample_to_isotropic, load_image, save_image
from mipcandy.data.visualization import auto_convert, visualize2d, visualize3d, overlay



from abc import ABCMeta, abstractmethod
from math import log, ceil
from os import PathLike, listdir
from os.path import isdir, basename
from typing import Sequence, Mapping, Any, override

import torch
from torch import nn

from mipcandy.common import Pad2d, Pad3d, Restore2d, Restore3d
from mipcandy.data import save_image, Loader, UnsupervisedDataset
from mipcandy.layer import WithPaddingModule
from mipcandy.sliding_window import SlidingWindow
from mipcandy.types import SupportedPredictant, Device


def parse_predictant(x: SupportedPredictant, loader: type[Loader], *, as_label: bool = False) -> tuple[list[
    torch.Tensor], list[str] | None]:
    if isinstance(x, str):
        if isdir(x):
            cases = listdir(x)
            return [loader.do_load(f"{x}/{case}", is_label=as_label) for case in cases], cases
        return [loader.do_load(x, is_label=as_label)], [basename(x)]
    if isinstance(x, torch.Tensor):
        return [x], None
    r, filenames = [], None
    for case in x:
        if isinstance(case, str):
            if not filenames:
                filenames = []
            r.append(loader.do_load(case, is_label=as_label))
            filenames.append(case[case.rfind("/") + 1:])
        elif filenames:
            raise TypeError("`x` should be single-typed")
        elif isinstance(case, torch.Tensor):
            r.append(case)
        else:
            raise TypeError(f"Unexpected type of element {type(case)}")
    return r, filenames


class Predictor(WithPaddingModule, metaclass=ABCMeta):
    def __init__(self, experiment_folder: str | PathLike[str], *, checkpoint: str = "checkpoint_best.pth",
                 device: Device = "cpu") -> None:
        super().__init__(device)
        self._experiment_folder: str = experiment_folder
        self._checkpoint: str = checkpoint
        self._model: nn.Module | None = None

    def lazy_load_model(self) -> None:
        if self._model:
            return
        self._model = self.build_network(torch.load(f"{self._experiment_folder}/{self._checkpoint}")).to(self._device)
        self._model.eval()

    @abstractmethod
    def build_network(self, checkpoint: Mapping[str, Any]) -> nn.Module:
        raise NotImplementedError

    def predict_image(self, image: torch.Tensor, *, batch: bool = False) -> torch.Tensor:
        self.lazy_load_model()
        image = image.to(self._device)
        if not batch:
            image = image.unsqueeze(0)
        padding_module = self.get_padding_module()
        if padding_module:
            image = padding_module(image)
        mask = self._model(image)
        restoring_module = self.get_restoring_module()
        if restoring_module:
            mask = restoring_module(mask)
        return mask if batch else mask.squeeze(0)

    def _predict(self, x: SupportedPredictant | UnsupervisedDataset) -> tuple[list[torch.Tensor], list[str] | None]:
        if isinstance(x, UnsupervisedDataset):
            return [self.predict_image(case) for case in x], None
        images, filenames = parse_predictant(x, Loader)
        return [self.predict_image(image) for image in images], filenames

    def predict(self, x: SupportedPredictant | UnsupervisedDataset) -> list[torch.Tensor]:
        return self._predict(x)[0]

    @staticmethod
    def save_prediction(mask: torch.Tensor, path: str | PathLike[str]) -> None:
        save_image(mask, path)

    def save_predictions(self, masks: Sequence[torch.Tensor], folder: str | PathLike[str], *,
                         filenames: Sequence[str | PathLike[str]] | None = None) -> None:
        if not filenames:
            num_cases = len(masks)
            num_digits = ceil(log(num_cases))
            filenames = [f"prediction_{str(i).zfill(num_digits)}" for i in range(num_cases)]
        for i, prediction in enumerate(masks):
            self.save_prediction(prediction, f"{folder}/{filenames[i]}")

    def predict_to_files(self, x: SupportedPredictant | UnsupervisedDataset, folder: str | PathLike[str]) -> None:
        masks, filenames = self._predict(x)
        self.save_predictions(masks, folder, filenames=filenames)

    def __call__(self, x: SupportedPredictant | UnsupervisedDataset) -> list[torch.Tensor]:
        return self.predict(x)


class SlidingPredictor(Predictor, SlidingWindow, metaclass=ABCMeta):
    @override
    def build_padding_module(self) -> nn.Module | None:
        window_shape = self.get_window_shape()
        return (Pad2d if len(window_shape) == 2 else Pad3d)(window_shape)

    @override
    def build_restoring_module(self, padding_module: nn.Module | None) -> nn.Module | None:
        if not isinstance(padding_module, (Pad2d, Pad3d)):
            raise TypeError("`padding_module` should be either `Pad2d` or `Pad3d`")
        window_shape = self.get_window_shape()
        return (Restore2d if len(window_shape) == 2 else Restore3d)(padding_module)

    @override
    def predict_image(self, image: torch.Tensor, *, batch: bool = False) -> torch.Tensor:
        if not batch:
            image = image.unsqueeze(0)
        images, metadata = self.do_sliding_window(image)
        masks = super().predict_image(images, batch=True)
        masks = self.revert_sliding_window(masks, metadata)
        return masks if batch else masks.squeeze(0)



from typing import Protocol, Literal

import torch

from mipcandy.types import Device

try:
    from cupy import from_dlpack as _dlpack2np
    from cupyx.scipy.ndimage import distance_transform_edt as _distance_transform_edt
except ImportError:
    from numpy import from_dlpack as _dlpack2np
    from scipy.ndimage import distance_transform_edt as _distance_transform_edt


def _args_check(mask: torch.Tensor, label: torch.Tensor, *, dtype: torch.dtype | None = None,
                device: Device | None = None) -> tuple[torch.dtype, Device]:
    if mask.shape != label.shape:
        raise ValueError(f"Mask ({mask.shape}) and label ({label.shape}) must have the same shape")
    if (mask_dtype := mask.dtype) != label.dtype or dtype and mask_dtype != dtype:
        raise TypeError(f"Mask ({mask_dtype}) and label ({label.dtype}) must both be {dtype}")
    if (mask_device := mask.device) != label.device:
        raise RuntimeError(f"Mask ({mask.device}) and label ({label.device}) must be on the same device")
    if device and mask_device != device:
        raise RuntimeError(f"Tensors are expected to be on {device}, but instead they are on {mask.device}")
    return mask_dtype, mask_device


class Metric(Protocol):
    def __call__(self, mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = ...) -> torch.Tensor: ...


def do_reduction(x: torch.Tensor, method: Literal["mean", "median", "sum", "none"] = "mean") -> torch.Tensor:
    match method:
        case "mean":
            return x.mean()
        case "median":
            return x.median()
        case "sum":
            return x.sum()
        case "none":
            return x


def apply_multiclass_to_binary(metric: Metric, mask: torch.Tensor, label: torch.Tensor, num_classes: int | None,
                               if_empty: float, *, reduction: Literal["mean", "sum"] = "mean") -> torch.Tensor:
    _args_check(mask, label, dtype=torch.int)
    if not num_classes:
        num_classes = max(mask.max().item(), label.max().item())
    if num_classes == 0:
        return torch.tensor(if_empty, dtype=torch.float)
    else:
        x = torch.tensor([metric(mask == cls, label == cls, if_empty=if_empty) for cls in range(1, num_classes + 1)])
        return do_reduction(x, reduction)


def dice_similarity_coefficient_binary(mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = 1) -> torch.Tensor:
    _args_check(mask, label, dtype=torch.bool)
    volume_sum = mask.sum() + label.sum()
    if volume_sum == 0:
        return torch.tensor(if_empty, dtype=torch.float)
    return 2 * (mask & label).sum() / volume_sum


def dice_similarity_coefficient_multiclass(mask: torch.Tensor, label: torch.Tensor, *, num_classes: int | None = None,
                                           if_empty: float = 1) -> torch.Tensor:
    return apply_multiclass_to_binary(dice_similarity_coefficient_binary, mask, label, num_classes, if_empty)


def soft_dice_coefficient(mask: torch.Tensor, label: torch.Tensor, *, smooth: float = 1e-5) -> torch.Tensor:
    _args_check(mask, label)
    num = label.size(0)
    mask = mask.view(num, -1)
    label = label.view(num, -1)
    intersection = (mask * label)
    dice = (2 * intersection.sum(1) + smooth) / (mask.sum(1) + label.sum(1) + smooth)
    return dice.sum() / num


def accuracy_binary(mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = 1) -> torch.Tensor:
    _args_check(mask, label, dtype=torch.bool)
    numerator = (mask & label).sum() + (~mask & ~label).sum()
    denominator = numerator + (mask & ~label).sum() + (label & ~mask).sum()
    return torch.tensor(if_empty, dtype=torch.float) if denominator == 0 else numerator / denominator


def accuracy_multiclass(mask: torch.Tensor, label: torch.Tensor, *, num_classes: int | None = None,
                        if_empty: float = 1) -> torch.Tensor:
    return apply_multiclass_to_binary(accuracy_binary, mask, label, num_classes, if_empty)


def _precision_or_recall(mask: torch.Tensor, label: torch.Tensor, if_empty: float, is_precision: bool) -> torch.Tensor:
    _args_check(mask, label, dtype=torch.bool)
    tp = (mask & label).sum()
    denominator = mask.sum() if is_precision else label.sum()
    return torch.tensor(if_empty, dtype=torch.float) if denominator == 0 else tp / denominator


def precision_binary(mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = 1) -> torch.Tensor:
    return _precision_or_recall(mask, label, if_empty, True)


def precision_multiclass(mask: torch.Tensor, label: torch.Tensor, *, num_classes: int | None = None,
                         if_empty: float = 1) -> torch.Tensor:
    return apply_multiclass_to_binary(precision_binary, mask, label, num_classes, if_empty)


def recall_binary(mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = 1) -> torch.Tensor:
    return _precision_or_recall(mask, label, if_empty, False)


def recall_multiclass(mask: torch.Tensor, label: torch.Tensor, *, num_classes: int | None = None,
                      if_empty: float = 1) -> torch.Tensor:
    return apply_multiclass_to_binary(recall_binary, mask, label, num_classes, if_empty)


def iou_binary(mask: torch.Tensor, label: torch.Tensor, *, if_empty: float = 1) -> torch.Tensor:
    _args_check(mask, label, dtype=torch.bool)
    denominator = (mask | label).sum()
    return torch.tensor(if_empty, dtype=torch.float) if denominator == 0 else (mask & label).sum() / denominator


def iou_multiclass(mask: torch.Tensor, label: torch.Tensor, *, num_classes: int | None = None,
                   if_empty: float = 1) -> torch.Tensor:
    return apply_multiclass_to_binary(iou_binary, mask, label, num_classes, if_empty)



from dataclasses import dataclass
from io import StringIO
from typing import Sequence, override

import torch
from ptflops import get_model_complexity_info
from torch import nn

from mipcandy.types import Device


def num_trainable_params(model: nn.Module) -> int:
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def model_complexity_info(model: nn.Module, example_shape: Sequence[int]) -> tuple[float | None, float | None, str]:
    layer_stats = StringIO()
    macs, params = get_model_complexity_info(model, tuple(example_shape), ost=layer_stats, as_strings=False)
    return macs, params, layer_stats.getvalue()


@dataclass
class SanityCheckResult(object):
    num_macs: float
    num_params: float
    layer_stats: str
    output: torch.Tensor

    @override
    def __str__(self) -> str:
        return f"MACs: {self.num_macs / 1e9:.1f} G / Params: {self.num_params / 1e6:.1f} M"


def sanity_check(model: nn.Module, input_shape: Sequence[int], *, device: Device | None = None) -> SanityCheckResult:
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    num_macs, num_params, layer_stats = model_complexity_info(model, input_shape)
    if num_macs is None or num_params is None:
        raise RuntimeError("Failed to validate model")
    output = model.to(device).eval()(torch.randn(1, *input_shape, device=device))
    return SanityCheckResult(num_macs, num_params, layer_stats, output.squeeze(0))



from typing import override


from mipcandy.types import Setting, Settings


class Frontend(object):
    def __init__(self, secrets: Settings) -> None:
        self._secrets: Settings = secrets

    def require_nonempty_secret(self, entry: str, *, require_type: type | None = None) -> Setting:
        if entry not in self._secrets:
            raise ValueError(f"Missing secret {entry}")
        secret = self._secrets[entry]
        if require_type is None or isinstance(secret, require_type):
            return secret
        raise ValueError(f"Invalid secret type {type(secret)}, {require_type} expected")

    def on_experiment_created(self, experiment_id: str, trainer: str, model: str, note: str, num_params: float,
                              num_macs: float, num_epochs: int, early_stop_tolerance: int) -> None:
        ...

    def on_experiment_updated(self, experiment_id: str, epoch: int, metrics: dict[str, list[float]],
                              early_stop_tolerance: int) -> None:
        ...

    def on_experiment_completed(self, experiment_id: str) -> None:
        ...

    def on_experiment_interrupted(self, experiment_id: str, error: Exception) -> None:
        ...


def create_hybrid_frontend(*frontends: Frontend) -> type[Frontend]:
    class HybridFrontend(Frontend):
        def __init__(self, secrets: Settings) -> None:
            super().__init__(secrets)

        @override
        def on_experiment_created(self, experiment_id: str, trainer: str, model: str, note: str, num_macs: float,
                                  num_params: float, num_epochs: int, early_stop_tolerance: int) -> None:
            for frontend in frontends:
                frontend.on_experiment_created(experiment_id, trainer, model, note, num_macs, num_params, num_epochs,
                                               early_stop_tolerance)

        @override
        def on_experiment_updated(self, experiment_id: str, epoch: int, metrics: dict[str, list[float]],
                                  early_stop_tolerance: int) -> None:
            for frontend in frontends:
                frontend.on_experiment_updated(experiment_id, epoch, metrics, early_stop_tolerance)

        @override
        def on_experiment_completed(self, experiment_id: str) -> None:
            for frontend in frontends:
                frontend.on_experiment_completed(experiment_id)

        @override
        def on_experiment_interrupted(self, experiment_id: str, error: Exception) -> None:
            for frontend in frontends:
                frontend.on_experiment_interrupted(experiment_id, error)

    return HybridFrontend



from typing import override

from wandb import init, Run

from mipcandy.frontend.prototype import Frontend
from mipcandy.types import Settings


class WandBFrontend(Frontend):
    def __init__(self, secrets: Settings) -> None:
        super().__init__(secrets)
        self._entity: str = self.require_nonempty_secret("wandb_entity", require_type=str)
        self._project: str = self.require_nonempty_secret("wandb_project", require_type=str)
        self._run: Run | None = None

    @override
    def on_experiment_created(self, experiment_id: str, trainer: str, model: str, note: str, num_macs: float,
                              num_params: float, num_epochs: int, early_stop_tolerance: int) -> None:
        self._run = init(entity=self._entity, project=self._project, config={
            "experiment_id": experiment_id, "trainer": trainer, "model": model, "note": note, "num_macs": num_macs,
            "num_params": num_params, "num_epochs": num_epochs
        })

    @override
    def on_experiment_updated(self, experiment_id: str, epoch: int, metrics: dict[str, list[float]],
                              early_stop_tolerance: int) -> None:
        if self._run:
            self._run.log(metrics)

    @override
    def on_experiment_completed(self, experiment_id: str) -> None:
        if not self._run:
            raise RuntimeError("Experiment has not been created")
        self._run.finish()
